package org.jboss.gravel.data.action;

import java.io.Serializable;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;
import org.jboss.gravel.Filter;

import javax.el.ELContext;
import javax.el.ELResolver;
import javax.el.FunctionMapper;
import javax.el.ValueExpression;
import javax.el.VariableMapper;
import javax.faces.context.FacesContext;
import javax.faces.event.AbortProcessingException;
import javax.faces.event.ActionEvent;
import javax.faces.event.ActionListener;

import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 */
public final class FilterActionListener implements ActionListener, Serializable {
    private final ValueExpression sourceExpression;
    private final ValueExpression targetExpression;
    private final String entryVarName;
    private final ValueExpression testExpression;
    private final ValueExpression filterExpression;
    private final ValueExpression filterArgumentExpression;
    private final ValueExpression ifExpression;

    private static final long serialVersionUID = 1L;

    public FilterActionListener(final ValueExpression sourceExpression, final ValueExpression targetExpression, final String entryVarName, final ValueExpression testExpression, final ValueExpression filterExpression, final ValueExpression filterArgumentExpression, final ValueExpression ifExpression) {
        this.sourceExpression = sourceExpression;
        this.targetExpression = targetExpression;
        this.entryVarName = entryVarName;
        this.testExpression = testExpression;
        this.filterExpression = filterExpression;
        this.filterArgumentExpression = filterArgumentExpression;
        this.ifExpression = ifExpression;
        if (log.isLoggable(Level.FINE)) {
            log.fine("Created new filter action listener: entry var is '" + entryVarName + "'");
        }
    }

    @SuppressWarnings("unchecked")
    public void processAction(ActionEvent event) throws AbortProcessingException {
        FacesContext facesContext = FacesContext.getCurrentInstance();
        final ELContext elContext = facesContext.getELContext();
        if (ifExpression != null) {
            Object ifValue = ifExpression.getValue(elContext);
            if (! Boolean.TRUE.equals(ifValue)) {
                return;
            }
        }
        final EntryELContext entryElContext = new EntryELContext(entryVarName, elContext);
        final Object sourceValue = sourceExpression.getValue(elContext);
        if (sourceValue == null) {
            targetExpression.setValue(elContext, null);
            return;
        }
        if (sourceValue instanceof SortedMap) {
            SortedMap<?, ?> sourceSortedMap = (SortedMap<?, ?>) sourceValue;
            SortedMap<Object, Object> targetSortedMap = new TreeMap<Object, Object>((Comparator<? super Object>)sourceSortedMap.comparator());
            for (Map.Entry<?,?> entry : sourceSortedMap.entrySet()) {
                if (checkItem(entryElContext, entry)) {
                    targetSortedMap.put(entry.getKey(), entry.getValue());
                }
            }
            targetExpression.setValue(elContext, Collections.unmodifiableSortedMap(targetSortedMap));
        } else if (sourceValue instanceof Map) {
            Map<?, ?> sourceMap = (Map<?, ?>) sourceValue;
            Map<Object, Object> targetMap = new HashMap<Object, Object>();
            for (Map.Entry<?,?> entry : sourceMap.entrySet()) {
                if (checkItem(entryElContext, entry)) {
                    targetMap.put(entry.getKey(), entry.getValue());
                }
            }
            targetExpression.setValue(elContext, Collections.unmodifiableMap(targetMap));
        } else if (sourceValue instanceof SortedSet) {
            SortedSet<?> sourceSet = (SortedSet<?>) sourceValue;
            SortedSet<Object> targetSet = new TreeSet<Object>((Comparator<? super Object>)sourceSet.comparator());
            for (Object entry : sourceSet) {
                if (checkItem(entryElContext, entry)) {
                    targetSet.add(entry);
                }
            }
            targetExpression.setValue(elContext, Collections.unmodifiableSortedSet(targetSet));
        } else if (sourceValue instanceof Set) {
            Set<?> sourceSet = (Set<?>) sourceValue;
            Set<Object> targetSet = new HashSet<Object>();
            for (Object entry : sourceSet) {
                if (checkItem(entryElContext, entry)) {
                    targetSet.add(entry);
                }
            }
            targetExpression.setValue(elContext, Collections.unmodifiableSet(targetSet));
        } else if (sourceValue instanceof List) {
            List<?> sourceList = (List<?>) sourceValue;
            List<Object> targetList = new ArrayList<Object>();
            for (Object entry : sourceList) {
                if (checkItem(entryElContext, entry)) {
                    targetList.add(entry);
                }
            }
            targetExpression.setValue(elContext, Collections.unmodifiableList(targetList));
        } else if (sourceValue instanceof Collection) {
            Collection<?> sourceCollection = (Collection<?>) sourceValue;
            Collection<Object> targetCollection = new ArrayList<Object>();
            for (Object entry : sourceCollection) {
                if (checkItem(entryElContext, entry)) {
                    targetCollection.add(entry);
                }
            }
            targetExpression.setValue(elContext, Collections.unmodifiableCollection(targetCollection));
        } else if (sourceValue.getClass().isArray()) {
            final Class<? extends Object> sourceValueClass = sourceValue.getClass();
            final int sourceLength = Array.getLength(sourceValue);
            Object targetArray = Array.newInstance(sourceValueClass, sourceLength);
            int srcIdx, targetIdx;
            for (srcIdx = 0, targetIdx = 0; srcIdx < sourceLength; srcIdx ++) {
                final Object value = Array.get(sourceValue, srcIdx);
                if (checkItem(entryElContext, value)) {
                    Array.set(targetArray, targetIdx++, value);
                }
            }
            if (targetIdx < sourceLength) {
                final Object newTargetArray = Array.newInstance(sourceValueClass, targetIdx);
                System.arraycopy(targetArray, 0, newTargetArray, 0, targetIdx);
                targetArray = newTargetArray;
            }
            targetExpression.setValue(elContext, targetArray);
        } else {
            if (checkItem(entryElContext, sourceValue)) {
                targetExpression.setValue(elContext, sourceValue);
            } else {
                targetExpression.setValue(elContext, null);
            }
        }
    }

    private boolean checkItem(final EntryELContext entryElContext, final Object item) {
        final boolean result;
        entryElContext.setItem(item);
        if (testExpression != null) {
            final Object testValue = testExpression.getValue(entryElContext);
            if (testValue != null) {
                if (testValue instanceof Boolean) {
                    result = ((Boolean)testValue).booleanValue();
                } else {
                    result = Boolean.valueOf(testValue.toString());
                }
            } else {
                result = false;
            }
            return result;
        } else if (filterExpression != null) {
            final Filter filterValue = (Filter) filterExpression.getValue(entryElContext);
            if (filterValue != null) {
                final Object filterArgument;
                if (filterArgumentExpression != null) {
                    filterArgument = filterArgumentExpression.getValue(entryElContext);
                } else {
                    filterArgument = item;
                }
                return filterValue.testEntry(filterArgument);
            } else {
                return true;
            }
        } else {
            return true;
        }
    }

    private static final class EntryELContext extends ELContext {
        private final ELContext elContext;
        private final EntryELResolver elResolver;

        public EntryELContext(final String entryVarName, final ELContext elContext) {
            this.elContext = elContext;
            elResolver = new EntryELResolver(entryVarName, elContext.getELResolver());
        }

        public void setItem(final Object item) {
            elResolver.setItem(item);
        }

        public void setPropertyResolved(boolean b) {
            elContext.setPropertyResolved(b);
        }

        public boolean isPropertyResolved() {
            return elContext.isPropertyResolved();
        }

        public void putContext(Class aClass, Object object) {
            elContext.putContext(aClass, object);
        }

        public Object getContext(Class aClass) {
            return elContext.getContext(aClass);
        }

        public ELResolver getELResolver() {
            return elResolver;
        }

        public FunctionMapper getFunctionMapper() {
            return elContext.getFunctionMapper();
        }

        public Locale getLocale() {
            return elContext.getLocale();
        }

        public void setLocale(Locale locale) {
            elContext.setLocale(locale);
        }

        public VariableMapper getVariableMapper() {
            return elContext.getVariableMapper();
        }
    }

    private static final class EntryELResolver extends ELResolver {
        private Object item;

        private final String entryVarName;
        private final ELResolver elResolver;

        public EntryELResolver(final String entryVarName, final ELResolver elResolver) {
            this.entryVarName = entryVarName;
            this.elResolver = elResolver;
        }

        public Object getValue(ELContext elContext, Object base, Object property) {
            if (base == null && entryVarName.equals(property)) {
                return item;
            } else {
                return elResolver.getValue(elContext, base, property);
            }
        }

        public Class getType(ELContext elContext, Object base, Object property) {
            return elResolver.getType(elContext, base, property);
        }

        public void setValue(ELContext elContext, Object base, Object property, Object value) {
            elResolver.setValue(elContext, base, property, value);
        }

        public boolean isReadOnly(ELContext elContext, Object base, Object property) {
            return elResolver.isReadOnly(elContext, base, property);
        }

        public Iterator getFeatureDescriptors(ELContext elContext, Object base) {
            return elResolver.getFeatureDescriptors(elContext, base);
        }

        public Class getCommonPropertyType(ELContext elContext, Object base) {
            return elResolver.getCommonPropertyType(elContext, base);
        }

        public void setItem(final Object item) {
            this.item = item;
        }
    }

    private static final Logger log = Logger.getLogger("org.jboss.gravel.data.action.FilterActionListener");
}
