package org.jboss.gravel.data.action;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import javax.el.ELContext;
import javax.el.ELResolver;
import javax.el.FunctionMapper;
import javax.el.ValueExpression;
import javax.el.VariableMapper;
import javax.faces.context.FacesContext;
import javax.faces.event.AbortProcessingException;
import javax.faces.event.ActionEvent;
import javax.faces.event.ActionListener;

/**
 *
 */
public final class SortActionListener implements ActionListener, Serializable {
    private final ValueExpression sourceExpression;
    private final ValueExpression targetExpression;
    private final String entryVarName;
    private final ValueExpression comparatorExpression;
    private final ValueExpression argumentExpression;
    private final ValueExpression ifExpression;
    private final ValueExpression reverseExpression;

    private static final long serialVersionUID = 1L;

    public SortActionListener(final ValueExpression sourceExpression, final ValueExpression targetExpression, final String entryVarName, final ValueExpression comparatorExpression, final ValueExpression argumentExpression, final ValueExpression ifExpression, final ValueExpression reverseExpression) {
        this.sourceExpression = sourceExpression;
        this.targetExpression = targetExpression;
        this.entryVarName = entryVarName;
        this.comparatorExpression = comparatorExpression;
        this.argumentExpression = argumentExpression;
        this.ifExpression = ifExpression;
        this.reverseExpression = reverseExpression;
    }

    @SuppressWarnings("unchecked")
    public void processAction(ActionEvent event) throws AbortProcessingException {
        FacesContext facesContext = FacesContext.getCurrentInstance();
        final ELContext elContext = facesContext.getELContext();
        if (ifExpression != null) {
            Object ifValue = ifExpression.getValue(elContext);
            if (! Boolean.TRUE.equals(ifValue)) {
                return;
            }
        }
        final SortActionListener.EntryELContext entryElContext = new SortActionListener.EntryELContext(entryVarName, elContext);
        final Object sourceValue = sourceExpression.getValue(elContext);
        if (sourceValue == null) {
            targetExpression.setValue(elContext, null);
            return;
        }
        final Comparator comparator = new ELComparator(entryElContext);

        if (sourceValue instanceof Map) {
            final Map<?, ?> sourceMap = (Map<?, ?>) sourceValue;
            final Map<Object, Object> targetMap = new LinkedHashMap<Object, Object>();
            final Set<? extends Map.Entry<?, ?>> entrySet = sourceMap.entrySet();
            final List<? extends Map.Entry<?, ?>> entryList = new ArrayList<Map.Entry<?, ?>>(entrySet);
            Collections.sort(entryList, comparator);
            for (Map.Entry<?, ?> entry : entryList) {
                targetMap.put(entry.getKey(), entry.getValue());
            }
            targetExpression.setValue(elContext, Collections.unmodifiableMap(targetMap));
        } else if (sourceValue instanceof Set) {
            Set<?> sourceSet = (Set<?>) sourceValue;
            Set<Object> targetSet = new LinkedHashSet<Object>();
            final List<?> entryList = new ArrayList<Object>(sourceSet);
            Collections.sort(entryList, comparator);
            for (Object entry : entryList) {
                targetSet.add(entry);
            }
            targetExpression.setValue(elContext, Collections.unmodifiableSet(targetSet));
        } else if (sourceValue instanceof Collection) {
            Collection<?> sourceCollection = (Collection<?>) sourceValue;
            List<Object> targetList = new ArrayList<Object>(sourceCollection);
            Collections.sort(targetList, comparator);
            targetExpression.setValue(elContext, Collections.unmodifiableList(targetList));
        } else if (sourceValue instanceof Object[]) {
            Object[] sourceArray = (Object[]) sourceValue;
            final int length = sourceArray.length;
            Object[] targetArray = new Object[length];
            System.arraycopy(targetArray, 0, sourceArray, 0, length);
            Arrays.sort(targetArray, comparator);
            targetExpression.setValue(elContext, targetArray);
        } else {
            targetExpression.setValue(elContext, sourceValue);
        }
    }

    // This class is never serialized with the enclosing class
    @SuppressWarnings ({"ComparatorNotSerializable"})
    private final class ELComparator<T> implements Comparator<T> {
        private final EntryELContext entryELContext;
        private final Comparator<T> nestedComparator;
        private final boolean reverse;

        @SuppressWarnings ({"unchecked"})
        public ELComparator(final EntryELContext entryELContext) {
            this.entryELContext = entryELContext;
            nestedComparator = comparatorExpression == null ? null : (Comparator<T>)comparatorExpression.getValue(entryELContext);
            reverse = reverseExpression == null ? false : ((Boolean)reverseExpression.getValue(entryELContext)).booleanValue();
        }

        @SuppressWarnings ({"unchecked"})
        public int compare(final T o1, final T o2) {
            final T v1;
            final T v2;
            if (argumentExpression != null) {
                entryELContext.setItem(o1);
                v1 = (T) argumentExpression.getValue(entryELContext);
                entryELContext.setItem(o2);
                v2 = (T) argumentExpression.getValue(entryELContext);
            } else {
                v1 = o1;
                v2 = o2;
            }
            if (nestedComparator != null) {
                return reverse ? nestedComparator.compare(v2, v1) : nestedComparator.compare(v1, v2);
            } else {
                int res;
                if (v1 == null) {
                    if (v2 == null) {
                        return 0;
                    } else {
                        res = -((Comparable<T>)v2).compareTo(null);
                    }
                } else {
                    if (v2 == null) {
                        res = ((Comparable<T>)v1).compareTo(null);
                    } else {
                        res = ((Comparable<T>)v1).compareTo(v2);
                    }
                }
                return reverse ? -res : res;
            }
        }
    }

    private static final class EntryELContext extends ELContext {
        private final ELContext elContext;
        private final SortActionListener.EntryELResolver elResolver;

        public EntryELContext(final String entryVarName, final ELContext elContext) {
            this.elContext = elContext;
            elResolver = new SortActionListener.EntryELResolver(entryVarName, elContext.getELResolver());
        }

        public void setItem(final Object item) {
            elResolver.setItem(item);
        }

        public void setPropertyResolved(boolean b) {
            elContext.setPropertyResolved(b);
        }

        public boolean isPropertyResolved() {
            return elContext.isPropertyResolved();
        }

        public void putContext(Class aClass, Object object) {
            elContext.putContext(aClass, object);
        }

        public Object getContext(Class aClass) {
            return elContext.getContext(aClass);
        }

        public ELResolver getELResolver() {
            return elResolver;
        }

        public FunctionMapper getFunctionMapper() {
            return elContext.getFunctionMapper();
        }

        public Locale getLocale() {
            return elContext.getLocale();
        }

        public void setLocale(Locale locale) {
            elContext.setLocale(locale);
        }

        public VariableMapper getVariableMapper() {
            return elContext.getVariableMapper();
        }
    }

    private static final class EntryELResolver extends ELResolver {
        private Object item;

        private final String entryVarName;
        private final ELResolver elResolver;

        public EntryELResolver(final String entryVarName, final ELResolver elResolver) {
            this.entryVarName = entryVarName;
            this.elResolver = elResolver;
        }

        public Object getValue(ELContext elContext, Object base, Object property) {
            if (base == null && entryVarName != null && entryVarName.equals(property)) {
                return item;
            } else {
                return elResolver.getValue(elContext, base, property);
            }
        }

        public Class getType(ELContext elContext, Object base, Object property) {
            return elResolver.getType(elContext, base, property);
        }

        public void setValue(ELContext elContext, Object base, Object property, Object value) {
            elResolver.setValue(elContext, base, property, value);
        }

        public boolean isReadOnly(ELContext elContext, Object base, Object property) {
            return elResolver.isReadOnly(elContext, base, property);
        }

        public Iterator getFeatureDescriptors(ELContext elContext, Object base) {
            return elResolver.getFeatureDescriptors(elContext, base);
        }

        public Class getCommonPropertyType(ELContext elContext, Object base) {
            return elResolver.getCommonPropertyType(elContext, base);
        }

        public void setItem(final Object item) {
            this.item = item;
        }
    }
}
