package org.jboss.gravel.data.phase;

import java.io.IOException;
import java.io.InputStream;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import java.util.logging.Logger;

/**
 *
 */
public final class MimeParser {

    private final BoundaryDelimitedInputStream inputStream;

    public MimeParser(InputStream source, byte[] boundary) {
        inputStream = new BoundaryDelimitedInputStream(source, boundary);
    }

    private static final int BEGIN_FIRST = 0;

    private static final int BEGIN = 1;
    private static final int GOT_ONE = 2;
    private static final int END_LINE = 3;
    private static final int END_HEADERS = 4;
    private static final int INVALID_LINE = -1;

    private static final int INVALID_LINE_END = -2;
    private static final int INVALID_PART = -3;

    private static final Pattern dispositionPattern =
        Pattern.compile("^content-disposition:\\s*(\\S+)(?:;\\s+name\\s*=\\s*\"([^\"]*)\")?(?:;\\s+filename\\s*=\\s*\"([^\"]*)\")?", Pattern.CASE_INSENSITIVE);

    private static final Pattern contentTypePattern =
        Pattern.compile("^content-type:\\s*(.*)", Pattern.CASE_INSENSITIVE);

    private static final Pattern transferEncodingPattern =
        Pattern.compile("^content-transfer-encoding:\\s*(\\S+)", Pattern.CASE_INSENSITIVE);

    public <T> void parse(MimePartHandler<T> handler, T userData) throws IOException {
        ByteStringBuilder line = new ByteStringBuilder();
        while (! inputStream.isOuterStreamClosed()) {
            String name = "";
            String fileName = "";
            String contentType = "text/plain";
            String contentEncoding = "8bit";
            String disposition = null;
            int b = 0;
            int state = BEGIN_FIRST;
            OUT: do {
                b = inputStream.read();
                if (b >= 0) {
                    switch (state) {
                        case INVALID_LINE:
                            switch (b) {
                                case '\r':
                                    state = INVALID_LINE_END;
                                    break;
                            }
                            break;
                        case INVALID_LINE_END:
                            switch (b) {
                                case '\n':
                                    state = BEGIN;
                                    break;
                                case '\r':
                                    break;
                                default:
                                    state = INVALID_LINE;
                                    break;
                            }
                            break;
                        case INVALID_PART:
                            // consume all
                            break;
                        case BEGIN_FIRST:
                            switch (b) {
                                case '\r':
                                    state = INVALID_LINE_END;
                                    break;
                                default:
                                    line.append((byte)b);
                                    state = GOT_ONE;
                            }
                            break;
                        case BEGIN:
                            switch (b) {
                                case '\r':
                                    state = END_HEADERS;
                                    break;
                                default:
                                    line.append((byte)b);
                                    state = GOT_ONE;
                            }
                            break;
                        case GOT_ONE:
                            switch (b) {
                                case '\r':
                                    state = END_LINE;
                                    break;
                                default:
                                    line.append((byte)b);
                            }
                            break;
                        case END_LINE:
                            switch (b) {
                                case '\n':
                                    state = BEGIN;
                                    // XXX process line
                                    Matcher matcher;
                                    final String stringLine = line.toString("UTF-8");
                                    if ((matcher = dispositionPattern.matcher(stringLine)).matches()) {
                                        String s;
                                        disposition = matcher.group(1);
                                        s = matcher.group(2);
                                        if (s != null) {
                                            name = s;
                                        }
                                        s = matcher.group(3);
                                        if (s != null) {
                                            fileName = s;
                                        }
                                    } else if ((matcher = contentTypePattern.matcher(stringLine)).matches()) {
                                        contentType = matcher.group(1);
                                    } else if ((matcher = transferEncodingPattern.matcher(stringLine)).matches()) {
                                        contentEncoding = matcher.group(1);
                                    }
                                    line.clear();
                                    break;
                                default:
                                    log.fine("Junk after CR");
                                    state = INVALID_LINE;
                                    break;
                            }
                            break;
                        case END_HEADERS:
                            switch (b) {
                                case '\n':
                                    // XXX process stream
                                    if (disposition != null) {
                                        final SectionInputStream dataStream = new SectionInputStream(inputStream);
                                        handler.handlePart(userData, name, fileName, contentType, contentEncoding, disposition, dataStream);
                                        while (dataStream.read() > -1);
                                    } else {
                                        log.fine("Skipping empty section");
                                    }
                                    break OUT;
                                default:
                                    log.fine("Junk after CR at end of headers");
                                    line.clear();
                                    state = INVALID_PART;
                                    break;
                            }
                            break;
                    }
                }
            } while (b != -1);
            line.clear();
        }
    }

    /**
     * Only reads eof from inner stream once; eats CR/LF at end of stream
     */
    private static final class SectionInputStream extends InputStream {
        private final InputStream inner;
        private static final int START = 0;
        private static final int GOT_CR = 1;
        private static final int GOT_CRLF = 2;
        private static final int EOF = 3;
        private static final int PLAYBACK_ONE = 4;
        private static final int PLAYBACK_LF_ONE = 5;
        private int state = START;
        private int next;

        public SectionInputStream(final InputStream inner) {
            this.inner = inner;
        }

        public int read() throws IOException {
            int r;
            switch (state) {
                case EOF:
                    return -1;
                case START:
                    r = inner.read();
                    switch (r) {
                        case -1:
                            state = EOF;
                            return -1;
                        case 13:
                            state = GOT_CR;
                            return read();
                        default:
                            return r;
                    }
                case GOT_CR:
                    r = inner.read();
                    switch (r) {
                        case -1:
                            state = EOF;
                            return 13;
                        case 10:
                            state = GOT_CRLF;
                            return read();
                        default:
                            next = r;
                            state = PLAYBACK_ONE;
                            return 13;
                    }
                case GOT_CRLF:
                    r = inner.read();
                    switch (r) {
                        case -1:
                            state = EOF;
                            return -1;
                        default:
                            next = r;
                            state = PLAYBACK_LF_ONE;
                            return 13;
                    }
                case PLAYBACK_ONE:
                    state = START;
                    return next;
                case PLAYBACK_LF_ONE:
                    state = PLAYBACK_ONE;
                    return 10;
                default:
                    throw new IllegalStateException("Illegal state");
            }
        }

        public long skip(long n) throws IOException {
            return inner.skip(n);
        }

        public boolean markSupported() {
            return false;
        }

        public int available() throws IOException {
            return 0;
        }
    }

    private static final Logger log = Logger.getLogger("org.jboss.gravel.data.phase.MimeParser");
}
