package org.jboss.gravel.data.renderer;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.List;
import org.jboss.gravel.common.phase.GravelPhaseListener;
import org.jboss.gravel.common.renderer.Element;
import org.jboss.gravel.common.renderer.RendererBase;
import org.jboss.gravel.data.ui.UIInputFile;

import javax.faces.FacesException;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;

import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 */
public final class InputFileRenderer extends RendererBase {
    private final long maxSize;

    public InputFileRenderer(final long maxSize) {
        this.maxSize = maxSize;
    }

    public void decode(FacesContext context, UIComponent component) {
        if (! component.isRendered()) {
            return;
        }
        final UIInputFile fileInput = (UIInputFile) component;
        final String fileKey = "gravel.private.file." + fileInput.getClientId(context);
        final File file = (File) context.getExternalContext().getRequestMap().get(fileKey);
        if (file != null) {
            final UIInputFile.Type type = fileInput.getTargetTypeEnum();
            switch (type) {
                case BYTES:
                    handleBytes(file, fileInput);
                    break;
                case CHARS:
                    handleChars(file, fileInput);
                    break;
                case READER:
                    handleReader(context, file, fileInput);
                    break;
                case STREAM:
                    handleStream(context, file, fileInput);
                    break;
                case STRING:
                    handleString(file, fileInput);
                    break;
            }
        }
    }

    private void handleString(final File file, final UIInputFile fileInput) {
        final long length = file.length();
        final long componentMaxSize = fileInput.getMaxSize();
        final long maxSize = componentMaxSize == -1 ? this.maxSize : componentMaxSize;
        final int size =
            (length > maxSize && maxSize > (long)Integer.MAX_VALUE) ||
            (length > (long)Integer.MAX_VALUE) ?
                Integer.MAX_VALUE :
            (length > maxSize) ?
                (int) maxSize :
                (int) length;
        final String charsetName = fileInput.getTargetCharset();
        char[] data = new char[size];
        try {
            final FileInputStream inputStream = new FileInputStream(file);
            boolean inputStreamOk = false;
            try {
                final Reader reader = new InputStreamReader(inputStream, charsetName == null ? "UTF-8" : charsetName);
                boolean readerOk = false;
                try {
                    int res, cnt = 0;
                    do {
                        res = reader.read(data, cnt, data.length - cnt);
                        if (res > 0) {
                            cnt += res;
                        }
                    } while (res != -1 && cnt < data.length);
                    reader.close();
                    readerOk = true;
                } finally {
                    if (!readerOk) {
                        try {
                            reader.close();
                        } catch (IOException ex) {
                            log.log(Level.FINE, "Error closing resource on error path", ex);
                        }
                    }
                }
                inputStream.close();
                inputStreamOk = true;
            } finally {
                if (!inputStreamOk) {
                    try {
                        inputStream.close();
                    } catch (IOException ex) {
                        log.log(Level.FINE, "Error closing resource on error path", ex);
                    }
                }
            }
        } catch (IOException e) {
            FacesException fe = new FacesException("Error reading file upload data: " + e.getMessage());
            fe.setStackTrace(e.getStackTrace());
            throw fe;
        }
        fileInput.updateTarget(new String(data));
    }

    private void handleStream(final FacesContext context, final File file, final UIInputFile fileInput) {
        try {
            final FileInputStream inputStream = new FileInputStream(file);
            boolean inputStreamOk = false;
            try {
                fileInput.updateTarget(inputStream);
                final List<Runnable> cleaners = GravelPhaseListener.getRequestCleaners(context);
                cleaners.add(GravelPhaseListener.closeableCleaner(inputStream));
                inputStreamOk = true;
            } finally {
                if (!inputStreamOk) {
                    try {
                        inputStream.close();
                    } catch (IOException ex) {
                        log.log(Level.FINE, "Error closing resource on error path", ex);
                    }
                }
            }
        } catch (IOException e) {
            FacesException fe = new FacesException("Error reading file upload data: " + e.getMessage());
            fe.setStackTrace(e.getStackTrace());
            throw fe;
        }
    }

    private void handleReader(final FacesContext context, final File file, final UIInputFile fileInput) {
        final String charsetName = fileInput.getTargetCharset();
        try {
            final FileInputStream inputStream = new FileInputStream(file);
            boolean inputStreamOk = false;
            try {
                final Reader reader = new InputStreamReader(inputStream, charsetName == null ? "UTF-8" : charsetName);
                boolean readerOk = false;
                try {
                    final List<Runnable> cleaners = GravelPhaseListener.getRequestCleaners(context);
                    cleaners.add(GravelPhaseListener.closeableCleaner(reader));
                    cleaners.add(GravelPhaseListener.closeableCleaner(inputStream));
                    fileInput.updateTarget(reader);
                    readerOk = true;
                } finally {
                    if (!readerOk) {
                        try {
                            reader.close();
                        } catch (IOException ex) {
                            log.log(Level.FINE, "Error closing resource on error path", ex);
                        }
                    }
                }
                inputStreamOk = true;
            } finally {
                if (!inputStreamOk) {
                    try {
                        inputStream.close();
                    } catch (IOException ex) {
                        log.log(Level.FINE, "Error closing resource on error path", ex);
                    }
                }
            }
        } catch (IOException e) {
            FacesException fe = new FacesException("Error reading file upload data: " + e.getMessage());
            fe.setStackTrace(e.getStackTrace());
            throw fe;
        }
    }

    private void handleChars(final File file, final UIInputFile fileInput) {
        final long length = file.length();
        final long componentMaxSize = fileInput.getMaxSize();
        final long maxSize = componentMaxSize == -1 ? this.maxSize : componentMaxSize;
        final int size =
            (length > maxSize && maxSize > (long)Integer.MAX_VALUE) ||
            (length > (long)Integer.MAX_VALUE) ?
                Integer.MAX_VALUE :
            (length > maxSize) ?
                (int) maxSize :
                (int) length;
        final String charsetName = fileInput.getTargetCharset();
        char[] data = new char[size];
        try {
            final FileInputStream inputStream = new FileInputStream(file);
            boolean inputStreamOk = false;
            try {
                final Reader reader = new InputStreamReader(inputStream, charsetName == null ? "UTF-8" : charsetName);
                boolean readerOk = false;
                try {
                    int res, cnt = 0;
                    do {
                        res = reader.read(data, cnt, data.length - cnt);
                        if (res > 0) {
                            cnt += res;
                        }
                    } while (res != -1 && cnt < data.length);
                    reader.close();
                    readerOk = true;
                } finally {
                    if (!readerOk) {
                        try {
                            reader.close();
                        } catch (IOException ex) {
                            log.log(Level.FINE, "Error closing resource on error path", ex);
                        }
                    }
                }
                inputStream.close();
                inputStreamOk = true;
            } finally {
                if (!inputStreamOk) {
                    try {
                        inputStream.close();
                    } catch (IOException ex) {
                        log.log(Level.FINE, "Error closing resource on error path", ex);
                    }
                }
            }
        } catch (IOException e) {
            FacesException fe = new FacesException("Error reading file upload data: " + e.getMessage());
            fe.setStackTrace(e.getStackTrace());
            throw fe;
        }
        fileInput.updateTarget(data);
    }

    private void handleBytes(final File file, final UIInputFile fileInput) {
        final long length = file.length();
        final long componentMaxSize = fileInput.getMaxSize();
        final long maxSize = componentMaxSize == -1 ? this.maxSize : componentMaxSize;
        final int size =
            (length > maxSize && maxSize > (long)Integer.MAX_VALUE) ||
            (length > (long)Integer.MAX_VALUE) ?
                Integer.MAX_VALUE :
            (length > maxSize) ?
                (int) maxSize :
                (int) length;
        byte[] data = new byte[size];
        try {
            final FileInputStream inputStream = new FileInputStream(file);
            boolean inputStreamOk = false;
            try {
                int res, cnt = 0;
                do {
                    res = inputStream.read(data, cnt, data.length - cnt);
                    if (res > 0) {
                        cnt += res;
                    }
                } while (res != -1 && cnt < data.length);
                inputStream.close();
                inputStreamOk = true;
            } finally {
                if (!inputStreamOk) {
                    try {
                        inputStream.close();
                    } catch (IOException ex) {
                        log.log(Level.FINE, "Error closing resource on error path", ex);
                    }
                }
            }
        } catch (IOException e) {
            FacesException fe = new FacesException("Error reading file upload data: " + e.getMessage());
            fe.setStackTrace(e.getStackTrace());
            throw fe;
        }
        fileInput.updateTarget(data);
    }

    public void encodeBegin(FacesContext context, UIComponent component) throws IOException {
        // nothing
    }

    public void encodeEnd(FacesContext context, UIComponent component) throws IOException {
        if (! component.isRendered()) {
            return;
        }
        UIInputFile fileInput = (UIInputFile) component;
        final Element<UIComponent> input = writeElement(context, "input", component);
        input.writeAttribute("type", "file");
        input.writeId("id", false);
        input.writeId("name", false);
        writeHtmlBasic(input, fileInput);
        writeHtmlSelect(input, fileInput);
        input.writeAttribute("value", fileInput.getFileName());
        input.writeAttribute("disabled", fileInput.isDisabled());
        input.writeAttribute("readonly", fileInput.isReadonly());
        input.writeAttribute("size", fileInput.getSize());
        input.writeAttribute("maxlength", fileInput.getMaxlength());
        input.writeAttribute("alt", fileInput.getAlt());
        input.writeAttribute("tabindex", fileInput.getTabindex());
        input.writeAttribute("accesskey", fileInput.getAccesskey());
        input.writeAttribute("accept", fileInput.getAccept());
        input.close();
    }

    public boolean getRendersChildren() {
        return false;
    }

    private static final Logger log = Logger.getLogger("org.jboss.gravel.data.renderer.FileInputRenderer");
}
